<?php
/**
 * @package     Joomla.Plugin
 * @subpackage  System.updatenotification
 *
 * @copyright   Copyright (C) 2005 - 2016 Open Source Matters, Inc. All rights reserved.
 * @license     GNU General Public License version 2 or later; see LICENSE.txt
 */

defined('_JEXEC') or die;

// Uncomment the following line to enable debug mode (update notification email sent every single time)
// define('PLG_SYSTEM_EXTUPNOTIFICATION_DEBUG', 1);

/**
 * Joomla! Update Notification plugin
 *
 * Sends out an email to all Super Users or a predefined email address when a new Joomla! version is available.
 *
 * This plugin is a direct adaptation of the corresponding plugin in Akeeba Ltd's Admin Tools. The author has
 * consented to relicensing their plugin's code under GPLv2 or later (the original version was licensed under
 * GPLv3 or later) to allow its inclusion in the Joomla! CMS.
 *
 * @since  3.5
 */
class PlgSystemExtUpnotification extends JPlugin
{
	/**
	 * Load plugin language files automatically
	 *
	 * @var    boolean
	 * @since  3.6.3
	 */
	protected $autoloadLanguage = true;

	/**
	 * The update check and notification email code is triggered after the page has fully rendered.
	 *
	 * @return  void
	 *
	 * @since   3.5
	 */
	public function onAfterRender()
	{
		// Get the timeout for Joomla! updates, as configured in com_installer's component parameters
		JLoader::import('joomla.application.component.helper');
		$component = JComponentHelper::getComponent('com_installer');

		/** @var \Joomla\Registry\Registry $params */
		$params        = $component->params;
		$cache_timeout = (int) $params->get('cachetimeout', 6);
		
		// Calculate cache_timout depending on selected Sent Email interval paramater or use default setting from installer updates caching parameter.
		$sent_interval = (int) $this->params->get('report_after', 0);
		$cache_timeout = (!$sent_interval) ? (3600 * $cache_timeout) : (3600 * $sent_interval);
		
		// Do we need to run? Compare the last run timestamp stored in the plugin's options with the current
		// timestamp. If the difference is greater than the cache timeout we shall not execute again.
		$now  = time();
		$last = (int) $this->params->get('lastrun', 0);
		
		if (!defined('PLG_SYSTEM_EXTUPNOTIFICATION_DEBUG') && (abs($now - $last) < $cache_timeout))
		{
			return;
		}

		// Update last run status
		// If I have the time of the last run, I can update, otherwise insert
		$this->params->set('lastrun', $now);

		$db = JFactory::getDbo();
		$query = $db->getQuery(true)
					->update($db->qn('#__extensions'))
					->set($db->qn('params') . ' = ' . $db->q($this->params->toString('JSON')))
					->where($db->qn('type') . ' = ' . $db->q('plugin'))
					->where($db->qn('folder') . ' = ' . $db->q('system'))
					->where($db->qn('element') . ' = ' . $db->q('extupnotification'));

		try
		{
			// Lock the tables to prevent multiple plugin executions causing a race condition
			$db->lockTable('#__extensions');
		}
		catch (Exception $e)
		{
			// If we can't lock the tables it's too risky to continue execution
			return;
		}
		
		try
		{
			// Update the plugin parameters
			$result = $db->setQuery($query)->execute();

			$this->clearCacheGroups(array('com_plugins'), array(0, 1));
		}
		catch (Exception $exc)
		{
			// If we failed to execite
			$db->unlockTables();
			$result = false;
		}

		try
		{
			// Unlock the tables after writing
			$db->unlockTables();
		}
		catch (Exception $e)
		{
			// If we can't lock the tables assume we have somehow failed
			$result = false;
		}

		// Abort on failure
		if (!$result)
		{
			return;
		}

		// This is the extension ID to use for all Extensions
		$eid = 0; // Find all extensions

		// Get any available updates
		// Unfortunately Joomla! MVC doesn't allow us to autoload classes
		JModelLegacy::addIncludePath(JPATH_ADMINISTRATOR . '/components/com_installer/models', 'InstallerModel');

		// Get the update model and retrieve the Joomla! Extension updates
		$model = JModelLegacy::getInstance('update', 'InstallerModel');
		$model->setState('filter.extension_id', $eid);
      $updates = $model->getItems(); // Get Extension list of updates
      
		// If there are no updates we don't have to notify anyone about anything. This is NOT a duplicate check.
		if (empty($updates))
		{
			return;
		}
		
		// Check for any disabled Update Sites and store the ID's for it (to check later)
      $disabledUpdateSites = $model->getDisabledUpdateSites();
      $updateSites = array();
      $disabledUpdateSitesMsg = "";
		if ($disabledUpdateSites) {
	      $updateSites = $this->getAnyDisabledUpdateSites();
	      $disabledUpdateSitesMsg = JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_DISABLEDUPDATESITES_MSG'); // Generate notification message in email
      }

		// Generate array to use for table display of updateable extensions
		$data = array(); 
		foreach ($updates as $update)
		{
			if (!in_array("'".$update->update_site_id."'", $updateSites))
			{
	         $uType = ucfirst($update->type);
	         $data[] = array(  JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_HEADER_EXTENSION') => $update->name, 
	         						JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_HEADER_TYPE') => $uType, 
	         						JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_HEADER_CURRENTVERSION') => $update->current_version,
	         						JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_HEADER_LATESTVERSION') => $update->version,
	         						JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_HEADER_INFOURL') => $update->infourl
	         					);
	      }
			
		}
		
		// Check if $data is empty (because of disabled updates - they got removed). Do nothing if there's nothing to update.
		if (empty( $data ))
		{
			return;
		}

      $overview = "";
      // If disabledUpdateSites are found - add a message before the updatelist
      if (!empty($disabledUpdateSitesMsg))
         $overview .= "*** ".$disabledUpdateSitesMsg." ***\n";

      // Load the updateable extensions array into a ASCII table for the email to send
      $overview .= "<span style='font-family: Monaco, \"Lucida Console\", monospace; font-size: 10pt;'>";
         $overview .= str_replace(" ", "&nbsp;", $this->ascii_table( $data ));
      $overview .= "</span>";
      
		
		// If we're here, we have updates. First, get a link to the Extensions Update component.
		$baseURL  = JUri::base();
		$baseURL  = rtrim($baseURL, '/');
		$baseURL .= (substr($baseURL, -13) != 'administrator') ? '/administrator/' : '/';
		$baseURL .= 'index.php?option=com_installer&view=update';
		$uri      = new JUri($baseURL);

		/**
		 * Some third party security solutions require a secret query parameter to allow log in to the administrator
		 * backend of the site. The link generated above will be invalid and could probably block the user out of their
		 * site, confusing them (they can't understand the third party security solution is not part of Joomla! proper).
		 * So, we're calling the onBuildAdministratorLoginURL system plugin event to let these third party solutions
		 * add any necessary secret query parameters to the URL. The plugins are supposed to have a method with the
		 * signature:
		 *
		 * public function onBuildAdministratorLoginURL(JUri &$uri);
		 *
		 * The plugins should modify the $uri object directly and return null.
		 */

		JEventDispatcher::getInstance()->trigger('onBuildAdministratorLoginURL', array(&$uri));

		// Let's find out the email addresses to notify
		$superUsers    = array();
		$specificEmail = $this->params->get('email', '');



		if (!empty($specificEmail))
		{
			$superUsers = $this->getSuperUsers($specificEmail);
		}

		if (empty($superUsers))
		{
			$superUsers = $this->getSuperUsers();
		}

		if (empty($superUsers))
		{
			return;
		}


		/* Load the appropriate language. We try to load English (UK), the current user's language and the forced
		 * language preference, in this order. This ensures that we'll never end up with untranslated strings in the
		 * update email which would make Joomla! seem bad. So, please, if you don't fully understand what the
		 * following code does DO NOT TOUCH IT. It makes the difference between a hobbyist CMS and a professional
		 * solution! */
		$jLanguage = JFactory::getLanguage();
		$jLanguage->load('plg_system_extupnotification', JPATH_ADMINISTRATOR, 'en-GB', true, true);
		$jLanguage->load('plg_system_extupnotification', JPATH_ADMINISTRATOR, null, true, false);

		// Then try loading the preferred (forced) language
		$forcedLanguage = $this->params->get('language_override', '');

		if (!empty($forcedLanguage))
		{
			$jLanguage->load('plg_system_extupnotification', JPATH_ADMINISTRATOR, $forcedLanguage, true, false);
		}

		// Set up the email subject and body
		$email_subject = JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_EMAIL_SUBJECT');
		$email_body    = JText::_('PLG_SYSTEM_EXTUPNOTIFICATION_EMAIL_BODY');

		$jConfig  = JFactory::getConfig();
		$sitename = $jConfig->get('sitename');
		$mailFrom = $jConfig->get('mailfrom');
		$fromName = $jConfig->get('fromname');

		$substitutions = array (
         '[SITENAME]'   => $sitename,
			'[OVERVIEW]'	=> $overview,
         '[URL]'        => JUri::base(),
         '[LINK]'       => $uri->toString(),
         '\\n'          => "\n",
		);

		foreach ($substitutions as $k => $v)
		{
			$email_subject = str_replace($k, $v, $email_subject);
			$email_body    = str_replace($k, $v, $email_body);
		}
		// Replace /n for <br /> for correct HTML email display
      $email_body		= str_replace("\n", "<br />", $email_body); 

		// Send the emails to the Super Users
		foreach ($superUsers as $superUser)
		{
			$mailer = JFactory::getMailer();
         $mailer->isHtml(); // Set email to HTML format
			$mailer->setSender(array($mailFrom, $fromName));
			$mailer->addRecipient($superUser->email);
			$mailer->setSubject($email_subject);
			$mailer->setBody($email_body);
			$mailer->Send();
			
		}
	}

	/**
	 * Returns the Super Users email information. If you provide a comma separated $email list
	 * we will check that these emails do belong to Super Users and that they have not blocked
	 * system emails.
	 *
	 * @param   null|string  $email  A list of Super Users to email
	 *
	 * @return  array  The list of Super User emails
	 *
	 * @since   3.5
	 */
	private function getSuperUsers($email = null)
	{
		// Get a reference to the database object
		$db = JFactory::getDbo();

		// Convert the email list to an array
		if (!empty($email))
		{
			$temp   = explode(',', $email);
			$emails = array();

			foreach ($temp as $entry)
			{
				$entry    = trim($entry);
				$emails[] = $db->q($entry);
			}

			$emails = array_unique($emails);
		}
		else
		{
			$emails = array();
		}

		// Get a list of groups which have Super User privileges
		$ret = array();

		try
		{
			
	      $query = $db->getQuery(true)
						->select($db->qn('rules'))
						->from($db->qn('#__assets'))
						->where($db->qn('parent_id') . ' = ' . $db->q(0));
			$db->setQuery($query, 0, 1);
			$rulesJSON = $db->loadResult();
			$rules     = json_decode($rulesJSON, true);
			
			$rawGroups = $rules['core.admin'];
			$groups    = array();

			if (empty($rawGroups))
			{
				return $ret;
			}

			foreach ($rawGroups as $g => $enabled)
			{
				if ($enabled)
				{
					$groups[] = $db->q($g);
				}
			}

			if (empty($groups))
			{
				return $ret;
			}
		}
		catch (Exception $exc)
		{
			return $ret;
		}



		// Get the user IDs of users belonging to the SA groups
		try
		{
			$query = $db->getQuery(true)
						->select($db->qn('user_id'))
						->from($db->qn('#__user_usergroup_map'))
						->where($db->qn('group_id') . ' IN(' . implode(',', $groups) . ')');
			$db->setQuery($query);
			$rawUserIDs = $db->loadColumn(0);

			if (empty($rawUserIDs))
			{
				return $ret;
			}

			$userIDs = array();

			foreach ($rawUserIDs as $id)
			{
				$userIDs[] = $db->q($id);
			}
		}
		catch (Exception $exc)
		{
			return $ret;
		}

		// Get the user information for the Super Administrator users
		try
		{
			$query = $db->getQuery(true)
						->select(
							array(
								$db->qn('id'),
								$db->qn('username'),
								$db->qn('email'),
							)
						)->from($db->qn('#__users'))
						->where($db->qn('id') . ' IN(' . implode(',', $userIDs) . ')')
						->where($db->qn('sendEmail') . ' = ' . $db->q('1'));

			if (!empty($emails))
			{
				$query->where($db->qn('email') . 'IN(' . implode(',', $emails) . ')');
			}

			$db->setQuery($query);
			$ret = $db->loadObjectList();
		}
		catch (Exception $exc)
		{
			return $ret;
		}

		return $ret;
	}

	/**
	 * Clears cache groups. We use it to clear the plugins cache after we update the last run timestamp.
	 *
	 * @param   array  $clearGroups   The cache groups to clean
	 * @param   array  $cacheClients  The cache clients (site, admin) to clean
	 *
	 * @return  void
	 *
	 * @since   3.5
	 */
	private function clearCacheGroups(array $clearGroups, array $cacheClients = array(0, 1))
	{
		$conf = JFactory::getConfig();

		foreach ($clearGroups as $group)
		{
			foreach ($cacheClients as $client_id)
			{
				try
				{
					$options = array(
						'defaultgroup' => $group,
						'cachebase'    => ($client_id) ? JPATH_ADMINISTRATOR . '/cache' :
							$conf->get('cache_path', JPATH_SITE . '/cache')
					);

					$cache = JCache::getInstance('callback', $options);
					$cache->clean();
				}
				catch (Exception $e)
				{
					// Ignore it
				}
			}
		}
	}
	
   /**
    * Create ascii table out of an array (thanx StereoFrog).
    *
    * @param   array  $data   The array to use to create table
    * @return  ASCII formatted table with data out of array
    *
    * Array sample: $data = array(
    *							array('id' => 1, 'field1' => 'foo', 'field2' => 'bar'),
    *							array('id' => 2, 'field1' => 'lorem', 'field2' => 'ipsum'),
    *						); 
    *
    * Output table will be like
    *    .---------.---------.
    *    | HEAD1   | HEAD2   |
    *    :---------+---------:
    *    | Content | Line 1  |
    *    | Content | Line 2  |
    *    '---------'---------'
    */
	private function ascii_table($data) 
	{
           
		$keys = array_keys( end($data) );

		# calculate optimal width
		$wid = array_map('strlen', $keys);
		foreach($data as $row) 
		{
		    foreach( array_values($row) as $k => $v )
		        $wid[$k] = max( $wid[$k], strlen($v) );
		}

		# build format and separator strings
		foreach($wid as $k => $v) {
		    $fmt[$k] = "%-{$v}s";
		    $sept[$k] = str_repeat('-', $v);
		}
		
		$top = '.-' . implode('-.-', $sept) . '-.'; // Top line of table
		$sep = ':-' . implode('-+-', $sept) . '-:'; // Header titles seperator
      $fmt = '| ' . implode(' | ', $fmt) . ' |'; // Row formats
      $end = '\'-' . implode('-\'-', $sept) . '-\''; // End line of table

		# create header
		$buf = array($top, strtoupper(vsprintf($fmt, $keys)), $sep);

		# print data
		foreach($data as $row) {
		    $buf[] = vsprintf($fmt, $row);
		}
      $buf[] = $end;

		# finish
		return implode("\n", $buf);
	} 
	
	private function GetAnyDisabledUpdateSites()
	{
      $db = JFactory::getDbo();
      $ret = array();
      try
      {
         $query = $db->getQuery(true)
	            ->select('update_site_id')
   	         ->from($db->quoteName('#__update_sites'))
      	      ->where($db->quoteName('enabled') . ' = 0');
         $db->setQuery($query);
         $RawUpdateSiteIDs = $db->loadColumn(0);

         if (empty($RawUpdateSiteIDs))
         {
            return $ret;
         }

         $UpdateSiteIDs = array();

         foreach ($RawUpdateSiteIDs as $id)
         {
            $UpdateSiteIDs[] = $db->q($id);
         }

      	return $UpdateSiteIDs;

      }
      catch (Exception $exc)
      {
         return $ret;
      }
	}
}
